/*
 * Program:	IMAP transfer mailboxes
 *
 * Author:	Mark Crispin
 *		Networks and Distributed Computing
 *		Computing & Communications
 *		University of Washington
 *		Administration Building, AG-44
 *		Seattle, WA  98195
 *		Internet: MRC@CAC.Washington.EDU
 *
 * Date:	2 February 1994
 * Last Edited:	8 February 2000
 *
 * Copyright 2000 by the University of Washington
 *
 *  Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * above copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Washington not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  This software is made
 * available "as is", and
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
 * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
 * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <errno.h>
extern int errno;		/* just in case */
#include "mail.h"
#include "osdep.h"
#include "misc.h"
#include "linkage.h"
char *getpass ();		/* let's all do the cretin hop */

int critical = NIL;		/* flag saying in critical code */
int ddelim = -1;		/* destination delimiter */
FILE *f = NIL;


/* Function prototypes */

void msg_string_init (STRING *s,void *data,unsigned long size);
char msg_string_next (STRING *s);
void msg_string_setpos (STRING *s,unsigned long i);
int main (int argc,char *argv[]);
long imapxfer_append (MAILSTREAM *stream,void *data,char **flags,char **date,
		      STRING **message);

/* Append package */

typedef struct append_package {
  MAILSTREAM *stream;		/* source stream */
  unsigned long msgno;		/* current message number */
  char *flags;			/* current flags */
  char *date;			/* message internal date */
  STRING *message;		/* stringstruct of message */
} APPENDPACKAGE;


/* Message string driver for message stringstructs */

STRINGDRIVER msg_string = {
  msg_string_init,		/* initialize string structure */
  msg_string_next,		/* get next byte in string structure */
  msg_string_setpos		/* set position in string structure */
};

/* Initialize file string structure for file stringstruct
 * Accepts: string structure
 *	    pointer to message data structure
 *	    size of string
 */

void msg_string_init (STRING *s,void *data,unsigned long size)
{
  APPENDPACKAGE *md = (APPENDPACKAGE *) data;
  s->data = data;		/* note stream/msgno and header length */
  mail_fetchheader_full (md->stream,md->msgno,NIL,&s->data1,FT_PREFETCHTEXT);
#if 0
  s->size = size;		/* message size */
#else	/* This kludge is necessary because of broken IMAP servers (sigh!) */
  mail_fetchtext_full (md->stream,md->msgno,&s->size,NIL);
  s->size += s->data1;		/* header + body size */
#endif
  SETPOS (s,0);
}


/* Get next character from file stringstruct
 * Accepts: string structure
 * Returns: character, string structure chunk refreshed
 */

char msg_string_next (STRING *s)
{
  char c = *s->curpos++;	/* get next byte */
  SETPOS (s,GETPOS (s));	/* move to next chunk */
  return c;			/* return the byte */
}


/* Set string pointer position for file stringstruct
 * Accepts: string structure
 *	    new position
 */

void msg_string_setpos (STRING *s,unsigned long i)
{
  APPENDPACKAGE *md = (APPENDPACKAGE *) s->data;
  if (i < s->data1) {		/* want header? */
    s->chunk = mail_fetchheader (md->stream,md->msgno);
    s->chunksize = s->data1;	/* header length */
    s->offset = 0;		/* offset is start of message */
  }
  else if (i < s->size) {	/* want body */
    s->chunk = mail_fetchtext (md->stream,md->msgno);
    s->chunksize = s->size - s->data1;
    s->offset = s->data1;	/* offset is end of header */
  }
  else {			/* off end of message */
    s->chunk = NIL;		/* make sure that we crack on this then */
    s->chunksize = 1;		/* make sure SNX cracks the right way... */
    s->offset = i;
  }
				/* initial position and size */
  s->curpos = s->chunk + (i -= s->offset);
  s->cursize = s->chunksize - i;
}

/* Main program */

int main (int argc,char *argv[])
{
  int ret = 1;
  char *sp,*dp,*t,*t1;
  char c,tmp[MAILTMPLEN],dst[MAILTMPLEN];
  APPENDPACKAGE ap;
  STRING st;
  MAILSTREAM *stream;
#include "linkage.c"
				/* validate arguments */
  if ((argc != 3) || (argv[1][0] != '{') || !(sp = strchr (argv[1],'}')) ||
      (argv[2][0] != '{') || !(dp = strchr (argv[2],'}')))
    printf ("usage: %s \"{source_host}prefix\" \"{destination_host}prefix\"\n",
	    argv[0]);
				/* open source mailbox */
  else if (!(ap.stream = mail_open (NIL,argv[1],OP_HALFOPEN)))
    printf ("?Can't contact source server: %s\n",argv[1]);
				/* open destination server */
  else if (!(stream = mail_open (NIL,argv[2],OP_HALFOPEN))) {
    printf ("?Can't contact destination server %s\n",argv[2]);
    mail_close (ap.stream);
  }
  else if (!(f = tmpfile ())) {	/* open temporary file */
    puts ("?Can't open temporary file");
    mail_close (stream);
    mail_close (ap.stream);
  }
  else {
    puts ("Listing mailboxes...");
				/* get delimiter for destination prefix */
    strcpy (strchr (strcpy (tmp,stream->mailbox),'}') + 1,++dp);
    mail_list (ap.stream,tmp,"");
    rewind (f);			/* list all mailboxes matching prefix */
    if (ddelim < 0) {		/* if server failed to give delimiter */
      puts ("WARNING: unable to get destination hierarchy delimiter!");
      ddelim = 0;		/* default to none */
    }
    strcpy (strchr (strcpy (tmp,ap.stream->mailbox),'}') + 1,++sp);
    mail_list (ap.stream,tmp,"*");
    rewind (f);

				/* read back mailbox names */
    while (fgets (tmp,MAILTMPLEN-1,f)) {
      if (t = strchr (tmp+1,'\n')) *t = '\0';
      for (t = dst,t1 = stream->mailbox,c = NIL; (c != '}') && *t1;
	   *t++ = c= *t1++);
      for (t1 = dp; *t1; *t++ = *t1++);
				/* src and dst have different delimiters? */
      if (ddelim && (ddelim != tmp[0]))
				/* swap delimiters then */
	for (t1 = strchr (tmp+1,'}') + 1; c = *t1++; *t++ = c) {
	  if (c == ddelim) c = tmp[0] ? tmp[0] : 'x';
	  else if (c == tmp[0]) c = ddelim;
	}
				/* easy case */
      else for (t1 = strchr (tmp+1,'}') + 1; *t1; *t++ = *t1++);
      *t++ = '\0';
      ret = 1;			/* in case failure */
      printf ("Copying %s\n  => %s\n",tmp+1,dst);
      fflush (stdout);
      if (!(ap.stream = mail_open (ap.stream,tmp+1,OP_READONLY)))
	printf ("?Can't open source mailbox\n");
      else {
	printf ("Mailbox %s has %lu message(s).\n",ap.stream->mailbox,
		ap.stream->nmsgs);
				/* create destination mailbox */
	if (!mail_create (stream,dst))
	  printf ("WARNING: Can't create mailbox %s\n",dst);
	if (ap.stream->nmsgs) {
	  sprintf (tmp,"1:%lu",ap.stream->nmsgs);
				/* make sure we have all messages */
	  mail_fetchfast (ap.stream,tmp);
	  ap.msgno = 0;		/* prepare package */
	  ap.flags = ap.date = NIL;
	  ap.message = &st;
				/* copy messages */
	  if (mail_append_multiple (stream,dst,imapxfer_append,(void *) &ap)){
	    ret = 0;		/* this one copied OK */
	    puts ("[Ok]");	/* let user know */
	  }
	  else {		/* failed */
	    printf ("?Can't append message to mailbox %s -- aborting\n",dst);
	    break;
	  }
	}
      }
    }
    mail_close (stream);
    mail_close (ap.stream);	/* close off mailbox */
  }
  _exit (ret);
  return ret;			/* stupid gcc */
}

/* Append callback
 * Accepts: mail stream
 *	    append package
 *	    pointer to return flags
 *	    pointer to return date
 *	    pointer to return message stringstruct
 * Returns: T on success
 */

long imapxfer_append (MAILSTREAM *stream,void *data,char **flags,char **date,
		      STRING **message)
{
  char *t,*t1,tmp[MAILTMPLEN];
  unsigned long u;
  MESSAGECACHE *elt;
  APPENDPACKAGE *ap = (APPENDPACKAGE *) data;
  if (ap->flags) fs_give ((void **) &ap->flags);
  if (ap->date) fs_give ((void **) &ap->date);
  mail_gc (ap->stream,GC_TEXTS);
  if (++ap->msgno <= ap->stream->nmsgs) {
				/* initialize flag string */
    memset (t = tmp,0,MAILTMPLEN);
				/* output system flags */
    if ((elt = mail_elt (ap->stream,ap->msgno))->seen) strcat (t," \\Seen");
    if (elt->deleted) strcat (t," \\Deleted");
    if (elt->flagged) strcat (t," \\Flagged");
    if (elt->answered) strcat (t," \\Answered");
    if (elt->draft) strcat (t," \\Draft");
    if (u = elt->user_flags) do	/* any user flags? */
      if ((MAILTMPLEN - ((t += strlen (t)) - tmp)) >
	  (2 + strlen
	   (t1 = ap->stream->user_flags[find_rightmost_bit (&u)]))) {
	*t++ = ' ';		/* space delimiter */
	strcpy (t,t1);		/* copy the user flag */
      }
    while (u);			/* until no more user flags */
    *flags = ap->flags = cpystr (tmp + 1);
    *date = ap->date = cpystr (mail_date (tmp,elt));
    *message = ap->message;	/* message stringstruct */
    INIT (ap->message,msg_string,(void *) ap,elt->rfc822_size);
  }
  else *message = NIL;		/* all done */
  return LONGT;
}

/* Co-routines from MAIL library */


/* Message matches a search
 * Accepts: MAIL stream
 *	    message number
 */

void mm_searched (MAILSTREAM *stream,unsigned long msgno)
{
  fatal ("impossible mm_searched() call");
}


/* Message exists (i.e. there are that many messages in the mailbox)
 * Accepts: MAIL stream
 *	    message number
 */

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}


/* Message expunged
 * Accepts: MAIL stream
 *	    message number
 */

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
  fatal ("impossible mm_expunged() call");
}


/* Message flags update seen
 * Accepts: MAIL stream
 *	    message number
 */

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}

/* Mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_list (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* note destination delimiter */
  if (ddelim < 0) ddelim = delimiter;
				/* if got a selectable name */
  else if (!(attributes & LATT_NOSELECT) && *name)
    fprintf (f,"%c%s\n",delimiter,name);
}


/* Subscribe mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_lsub (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* dummy routine */
}


/* Mailbox status
 * Accepts: MAIL stream
 *	    mailbox name
 *	    mailbox status
 */

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
				/* dummy routine */
}

/* Notification event
 * Accepts: MAIL stream
 *	    string to log
 *	    error flag
 */

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
  mm_log (string,errflg);	/* just do mm_log action */
}


/* Log an event for the user to see
 * Accepts: string to log
 *	    error flag
 */

void mm_log (char *string,long errflg)
{
  switch (errflg) {  
  case BYE:
  case NIL:			/* no error */
    fprintf (stderr,"[%s]\n",string);
    break;
  case PARSE:			/* parsing problem */
  case WARN:			/* warning */
    fprintf (stderr,"%%%s\n",string);
    break;
  case ERROR:			/* error */
  default:
    fprintf (stderr,"?%s\n",string);
    break;
  }
}


/* Log an event to debugging telemetry
 * Accepts: string to log
 */

void mm_dlog (char *string)
{
  fprintf (stderr,"%s\n",string);
}

/* Get user name and password for this host
 * Accepts: parse of network mailbox name
 *	    where to return user name
 *	    where to return password
 *	    trial count
 */

void mm_login (NETMBX *mb,char *username,char *password,long trial)
{
  if (*mb->user) strcpy (username,mb->user);
  else {
    printf ("{%s/%s} username: ",mb->host,mb->service);
    gets (username);
  }
  strcpy (password,getpass ("password: "));
}


/* About to enter critical code
 * Accepts: stream
 */

void mm_critical (MAILSTREAM *stream)
{
  critical = T;			/* note in critical code */
}


/* About to exit critical code
 * Accepts: stream
 */

void mm_nocritical (MAILSTREAM *stream)
{
  critical = NIL;		/* note not in critical code */
}


/* Disk error found
 * Accepts: stream
 *	    system error code
 *	    flag indicating that mailbox may be clobbered
 * Returns: T if user wants to abort
 */

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
  return T;
}


/* Log a fatal error event
 * Accepts: string to log
 */

void mm_fatal (char *string)
{
  fprintf (stderr,"?%s\n",string);
}
