/*
 * Program:	Copy mailbox
 *
 * Author:	Mark Crispin
 *		Networks and Distributed Computing
 *		Computing & Communications
 *		University of Washington
 *		Administration Building, AG-44
 *		Seattle, WA  98195
 *		Internet: MRC@CAC.Washington.EDU
 *
 * Date:	28 July 1994
 * Last Edited:	2 January 2002
 *
 * The IMAP tools software provided in this Distribution is
 * Copyright 2002 University of Washington.
 * The full text of our legal notices is contained in the file called
 * CPYRIGHT, included with this Distribution.
 */

#include <stdio.h>
#include <errno.h>
extern int errno;		/* just in case */
#include "mail.h"
#include "osdep.h"
#include "misc.h"
#include "linkage.h"
char *getpass ();		/* let's all do the cretin hop */

int critical = NIL;		/* flag saying in critical code */
int debugp = NIL;		/* flag saying debugging */
int verbosep = NIL;		/* flag saying verbose */


/* Function prototypes */

int main (int argc,char *argv[]);

/* Main program */

int main (int argc,char *argv[])
{
  int ret = 1;
  char tmp[MAILTMPLEN];
  MAILSTREAM *s = NIL;
  char *pgm = argc ? argv[0] : "mbxcopy";
  char *pgx = strrchr (pgm,'/');
  int move = !strcmp (pgx ? pgx + 1 : pgm,"mbxmove");
#include "linkage.c"
				/* make sure correct number of arguments */
  if (argc != 3) printf ("usage: %s source destination\n",pgm);
				/* open source mailbox */
  else if (s = mail_open (s,argv[1],(move ? NIL : OP_READONLY))) {
    if (!mail_create (NIL,argv[2]))
      printf ("?Can't create mailbox %s\n",argv[2]);
				/* forget empty mailbox */
    else if (!s->nmsgs) printf ("Mailbox %s is empty\n",s->mailbox);
    else {
      printf ("%s [%lu message(s)] => %s\n",s->mailbox,s->nmsgs,argv[2]);
      sprintf (tmp,"1:%lu",s->nmsgs);
				/* copy messages */
      if (move ? mail_move (s,tmp,argv[2]) : mail_copy (s,tmp,argv[2])) {
	if (move) mail_expunge(s);/* flush moved messages */
	puts ("[Ok]");		/* make sure user knows it won */
	ret = NIL;
      }
      else puts ("[Failed]");
    }
    mail_close (s);		/* close off mailbox */
  }
  _exit (ret);
  return 0;			/* stupid gcc */
}

/* Co-routines from MAIL library */


/* Message matches a search
 * Accepts: MAIL stream
 *	    message number
 */

void mm_searched (MAILSTREAM *stream,unsigned long msgno)
{
  fatal ("impossible mm_searched() call");
}


/* Message exists (i.e. there are that many messages in the mailbox)
 * Accepts: MAIL stream
 *	    message number
 */

void mm_exists (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}


/* Message expunged
 * Accepts: MAIL stream
 *	    message number
 */

void mm_expunged (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}


/* Message flags update seen
 * Accepts: MAIL stream
 *	    message number
 */

void mm_flags (MAILSTREAM *stream,unsigned long number)
{
				/* dummy routine */
}

/* Mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_list (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* dummy routine */
}


/* Subscribe mailbox found
 * Accepts: MAIL stream
 *	    hierarchy delimiter
 *	    mailbox name
 *	    mailbox attributes
 */

void mm_lsub (MAILSTREAM *stream,int delimiter,char *name,long attributes)
{
				/* dummy routine */
}


/* Mailbox status
 * Accepts: MAIL stream
 *	    mailbox name
 *	    mailbox status
 */

void mm_status (MAILSTREAM *stream,char *mailbox,MAILSTATUS *status)
{
				/* dummy routine */
}

/* Notification event
 * Accepts: MAIL stream
 *	    string to log
 *	    error flag
 */

void mm_notify (MAILSTREAM *stream,char *string,long errflg)
{
  mm_log (string,errflg);	/* just do mm_log action */
}


/* Log an event for the user to see
 * Accepts: string to log
 *	    error flag
 */

void mm_log (char *string,long errflg)
{
  switch (errflg) {  
  case BYE:			/* server sayonara */
  case NIL:			/* no error */
    if (verbosep) printf ("[%s]\n",string);
    break;
  case PARSE:			/* parsing problem */
  case WARN:			/* warning */
    fprintf (stderr,"%%%s\n",string);
    break;
  case ERROR:			/* error */
  default:
    fprintf (stderr,"?%s\n",string);
    break;
  }
}


/* Log an event to debugging telemetry
 * Accepts: string to log
 */

void mm_dlog (char *string)
{
  fprintf (stderr,"%s\n",string);
}

/* Get user name and password for this host
 * Accepts: parse of network mailbox name
 *	    where to return user name
 *	    where to return password
 *	    trial count
 */

void mm_login (NETMBX *mb,char *username,char *password,long trial)
{
  if (*mb->user) strcpy (username,mb->user);
  else {
    printf ("{%s/%s} username: ",mb->host,mb->service);
    gets (username);
  }
  strcpy (password,getpass ("password: "));
}


/* About to enter critical code
 * Accepts: stream
 */

void mm_critical (MAILSTREAM *stream)
{
  critical = T;			/* note in critical code */
}


/* About to exit critical code
 * Accepts: stream
 */

void mm_nocritical (MAILSTREAM *stream)
{
  critical = NIL;		/* note not in critical code */
}


/* Disk error found
 * Accepts: stream
 *	    system error code
 *	    flag indicating that mailbox may be clobbered
 * Returns: T if user wants to abort
 */

long mm_diskerror (MAILSTREAM *stream,long errcode,long serious)
{
  return T;
}


/* Log a fatal error event
 * Accepts: string to log
 */

void mm_fatal (char *string)
{
  fprintf (stderr,"?%s\n",string);
}
